package blobovniczatree

import (
	"context"
	"encoding/hex"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobovnicza"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/util/logicerr"
	tracingPkg "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/tracing"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client"
	apistatus "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client/status"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
	"go.uber.org/zap"
)

// Delete deletes object from blobovnicza tree.
//
// If blobocvnicza ID is specified, only this blobovnicza is processed.
// Otherwise, all Blobovniczas are processed descending weight.
func (b *Blobovniczas) Delete(ctx context.Context, prm common.DeletePrm) (res common.DeleteRes, err error) {
	var (
		success   = false
		startedAt = time.Now()
	)
	defer func() {
		b.metrics.Delete(time.Since(startedAt), success, prm.StorageID != nil)
	}()

	ctx, span := tracing.StartSpanFromContext(ctx, "Blobovniczas.Delete",
		trace.WithAttributes(
			attribute.String("path", b.rootPath),
			attribute.String("address", prm.Address.EncodeToString()),
			attribute.String("storage_id", hex.EncodeToString(prm.StorageID)),
		))
	defer span.End()

	if b.readOnly {
		return common.DeleteRes{}, common.ErrReadOnly
	}

	var bPrm blobovnicza.DeletePrm
	bPrm.SetAddress(prm.Address)

	if prm.StorageID != nil {
		id := blobovnicza.NewIDFromBytes(prm.StorageID)
		shBlz := b.getBlobovnicza(id.String())
		blz, err := shBlz.Open()
		if err != nil {
			return res, err
		}
		defer shBlz.Close()

		if res, err = b.deleteObject(ctx, blz, bPrm); err == nil {
			success = true
		}
		return res, err
	}

	objectFound := false

	err = b.iterateSortedLeaves(ctx, &prm.Address, func(p string) (bool, error) {
		res, err = b.deleteObjectFromLevel(ctx, bPrm, p)
		if err != nil {
			if !client.IsErrObjectNotFound(err) {
				b.log.Debug(logs.BlobovniczatreeCouldNotRemoveObjectFromLevel,
					zap.String("level", p),
					zap.String("error", err.Error()),
					zap.String("trace_id", tracingPkg.GetTraceID(ctx)),
				)
			}
		}

		if err == nil {
			objectFound = true
		}

		// abort iterator if found, otherwise process all Blobovniczas
		return err == nil, nil
	})

	if err == nil && !objectFound {
		// not found in any blobovnicza
		return common.DeleteRes{}, logicerr.Wrap(new(apistatus.ObjectNotFound))
	}
	success = err == nil

	return
}

// tries to delete object from particular blobovnicza.
//
// returns no error if object was removed from some blobovnicza of the same level.
func (b *Blobovniczas) deleteObjectFromLevel(ctx context.Context, prm blobovnicza.DeletePrm, blzPath string) (common.DeleteRes, error) {
	shBlz := b.getBlobovnicza(blzPath)
	blz, err := shBlz.Open()
	if err != nil {
		return common.DeleteRes{}, err
	}
	defer shBlz.Close()

	return b.deleteObject(ctx, blz, prm)
}

// removes object from blobovnicza and returns common.DeleteRes.
func (b *Blobovniczas) deleteObject(ctx context.Context, blz *blobovnicza.Blobovnicza, prm blobovnicza.DeletePrm) (common.DeleteRes, error) {
	_, err := blz.Delete(ctx, prm)
	return common.DeleteRes{}, err
}
