package blobstor

import (
	"context"
	"encoding/hex"
	"strconv"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/util/logicerr"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client"
	apistatus "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client/status"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
)

// GetRange reads object payload data from b.
// If the descriptor is present, only one sub-storage is tried,
// Otherwise, each sub-storage is tried in order.
func (b *BlobStor) GetRange(ctx context.Context, prm common.GetRangePrm) (res common.GetRangeRes, err error) {
	startedAt := time.Now()
	defer func() {
		b.metrics.GetRange(time.Since(startedAt), len(res.Data), err == nil, prm.StorageID != nil)
	}()

	ctx, span := tracing.StartSpanFromContext(ctx, "BlobStor.GetRange",
		trace.WithAttributes(
			attribute.String("address", prm.Address.EncodeToString()),
			attribute.String("storage_id", hex.EncodeToString(prm.StorageID)),
			attribute.String("offset", strconv.FormatUint(prm.Range.GetOffset(), 10)),
			attribute.String("length", strconv.FormatUint(prm.Range.GetLength(), 10)),
		))
	defer span.End()

	b.modeMtx.RLock()
	defer b.modeMtx.RUnlock()

	if prm.StorageID == nil {
		for i := range b.storage {
			res, err = b.storage[i].Storage.GetRange(ctx, prm)
			if err == nil || !client.IsErrObjectNotFound(err) {
				return res, err
			}
		}

		return common.GetRangeRes{}, logicerr.Wrap(new(apistatus.ObjectNotFound))
	}
	if len(prm.StorageID) == 0 {
		res, err = b.storage[len(b.storage)-1].Storage.GetRange(ctx, prm)
	} else {
		res, err = b.storage[0].Storage.GetRange(ctx, prm)
	}
	return res, err
}
