package blobstor

import (
	"context"
	"fmt"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
	"go.uber.org/zap"
)

// Iterate traverses the storage over the stored objects and calls the handler
// on each element.
//
// Returns any error encountered that
// did not allow to completely iterate over the storage.
//
// If handler returns an error, method wraps and returns it immediately.
func (b *BlobStor) Iterate(ctx context.Context, prm common.IteratePrm) (common.IterateRes, error) {
	var (
		startedAt = time.Now()
		success   = false
	)
	defer func() {
		b.metrics.Iterate(time.Since(startedAt), success)
	}()
	ctx, span := tracing.StartSpanFromContext(ctx, "BlobStor.Iterate",
		trace.WithAttributes(
			attribute.Bool("ignore_errors", prm.IgnoreErrors),
		))
	defer span.End()

	b.modeMtx.RLock()
	defer b.modeMtx.RUnlock()

	for i := range b.storage {
		_, err := b.storage[i].Storage.Iterate(ctx, prm)
		if err != nil && !prm.IgnoreErrors {
			return common.IterateRes{}, fmt.Errorf("blobstor iterator failure: %w", err)
		}
	}
	success = true
	return common.IterateRes{}, nil
}

// IterateBinaryObjects is a helper function which iterates over BlobStor and passes binary objects to f.
// Errors related to object reading and unmarshaling are logged and skipped.
func IterateBinaryObjects(ctx context.Context, blz *BlobStor, f func(addr oid.Address, data []byte, descriptor []byte) error) error {
	var prm common.IteratePrm

	prm.Handler = func(elem common.IterationElement) error {
		return f(elem.Address, elem.ObjectData, elem.StorageID)
	}
	prm.IgnoreErrors = true
	prm.ErrorHandler = func(addr oid.Address, err error) error {
		blz.log.Warn(logs.BlobstorErrorOccurredDuringTheIteration,
			zap.Stringer("address", addr),
			zap.String("err", err.Error()))
		return nil
	}

	_, err := blz.Iterate(ctx, prm)

	return err
}
