package frostfs

import (
	"fmt"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/morph/client"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/morph/event"
	"github.com/nspcc-dev/neo-go/pkg/core/state"
	"github.com/nspcc-dev/neo-go/pkg/util"
)

// Deposit structure of frostfs.Deposit notification from mainnet chain.
type Deposit struct {
	IDValue     []byte
	AmountValue int64 // Fixed8
	FromValue   util.Uint160
	ToValue     util.Uint160
}

// MorphEvent implements Neo:Morph Event interface.
func (Deposit) MorphEvent() {}

// ID is a deposit transaction hash.
func (d Deposit) ID() []byte { return d.IDValue }

// From is a script hash of asset sender in main net.
func (d Deposit) From() util.Uint160 { return d.FromValue }

// To is a script hash of asset receiver in balance contract.
func (d Deposit) To() util.Uint160 { return d.ToValue }

// Amount of transferred assets.
func (d Deposit) Amount() int64 { return d.AmountValue }

// ParseDeposit notification into deposit structure.
func ParseDeposit(e *state.ContainedNotificationEvent) (event.Event, error) {
	var ev Deposit

	params, err := event.ParseStackArray(e)
	if err != nil {
		return nil, fmt.Errorf("could not parse stack items from notify event: %w", err)
	}

	if ln := len(params); ln != 4 {
		return nil, event.WrongNumberOfParameters(4, ln)
	}

	// parse from
	from, err := client.BytesFromStackItem(params[0])
	if err != nil {
		return nil, fmt.Errorf("could not get deposit sender: %w", err)
	}

	ev.FromValue, err = util.Uint160DecodeBytesBE(from)
	if err != nil {
		return nil, fmt.Errorf("could not convert deposit sender to uint160: %w", err)
	}

	// parse amount
	ev.AmountValue, err = client.IntFromStackItem(params[1])
	if err != nil {
		return nil, fmt.Errorf("could not get deposit amount: %w", err)
	}

	// parse to
	to, err := client.BytesFromStackItem(params[2])
	if err != nil {
		return nil, fmt.Errorf("could not get deposit receiver: %w", err)
	}

	ev.ToValue, err = util.Uint160DecodeBytesBE(to)
	if err != nil {
		return nil, fmt.Errorf("could not convert deposit receiver to uint160: %w", err)
	}

	// parse id
	ev.IDValue, err = client.BytesFromStackItem(params[3])
	if err != nil {
		return nil, fmt.Errorf("could not get deposit id: %w", err)
	}

	return ev, nil
}
